// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    bundle::{Bundle, BundleType},
    enum_::{Enum, EnumType},
    expr::{Expr, ExprEnum, ToExpr},
    int::{Bool, IntCmp, IntType, SInt, SIntValue, UInt, UIntValue, UnknownSize},
    intern::{Intern, Interned},
    ty::{CanonicalType, Type},
    util::GenericConstUsize,
};
use std::ops::{Add, BitAnd, BitOr, BitXor, Div, Mul, Neg as StdNeg, Not, Rem, Shl, Shr, Sub};

macro_rules! forward_value_to_expr_unary_op_trait {
    (
        #[generics($($generics:tt)*)]
        #[value($Value:ty)]
        $Trait:ident::$method:ident
    ) => {
        impl<$($generics)*> $Trait for $Value
        where
            Expr<<$Value as ToExpr>::Type>: $Trait,
        {
            type Output = <Expr<<$Value as ToExpr>::Type> as $Trait>::Output;

            fn $method(self) -> Self::Output {
                $Trait::$method(self.to_expr())
            }
        }
    };
}

macro_rules! impl_unary_op_trait {
    (
        #[generics($($generics:tt)*)]
        fn $Trait:ident::$method:ident($arg:ident: $Arg:ty) -> $Output:ty {
            $($body:tt)*
        }
    ) => {
        impl<$($generics)*> $Trait for Expr<$Arg>
        {
            type Output = Expr<$Output>;

            fn $method(self) -> Self::Output {
                let $arg = self;
                $($body)*
            }
        }
    };
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotU<Width: GenericConstUsize = UnknownSize> {
    pub arg: Expr<UInt<Width>>,
}

impl<Width: GenericConstUsize> ToExpr for NotU<Width> {
    type Type = UInt<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotU(NotU {
                arg: Expr::as_dyn_int(self.arg),
            })
            .intern(),
            __ty: self.arg.__ty,
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    fn Not::not(arg: UInt<Width>) -> UInt<Width> {
        NotU { arg }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    #[value(UIntValue<Width>)]
    Not::not
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotS<Width: GenericConstUsize = UnknownSize> {
    pub arg: Expr<SInt<Width>>,
}

impl<Width: GenericConstUsize> ToExpr for NotS<Width> {
    type Type = UInt<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotS(NotS {
                arg: Expr::as_dyn_int(self.arg),
            })
            .intern(),
            __ty: self.arg.__ty.as_same_width_uint(),
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    fn Not::not(arg: SInt<Width>) -> UInt<Width> {
        NotS { arg }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    #[value(SIntValue<Width>)]
    Not::not
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotB {
    pub arg: Expr<Bool>,
}

impl ToExpr for NotB {
    type Type = Bool;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotB(NotB { arg: self.arg }).intern(),
            __ty: self.arg.__ty,
        }
    }
}

impl_unary_op_trait! {
    #[generics()]
    fn Not::not(arg: Bool) -> Bool {
        NotB { arg }.to_expr()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Neg {
    pub arg: Expr<SInt>,
}

impl ToExpr for Neg {
    type Type = SInt;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::Neg(*self).intern(),
            __ty: SInt::new_dyn(
                Expr::ty(self.arg)
                    .width()
                    .checked_add(1)
                    .expect("width too big"),
            ),
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    fn StdNeg::neg(arg: SInt<Width>) -> SInt {
        Neg { arg: Expr::as_dyn_int(arg) }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    #[value(SIntValue<Width>)]
    StdNeg::neg
}

macro_rules! impl_binary_op_trait {
    (
        #[generics($($generics:tt)*)]
        fn $Trait:ident::$method:ident($lhs:ident: $Lhs:ty, $rhs:ident: $Rhs:ty) -> $Output:ty {
            $($body:tt)*
        }
    ) => {
        impl<
                Rhs: ToExpr<Type = $Rhs>,
                $($generics)*
            > $Trait<Rhs> for Expr<$Lhs>
        {
            type Output = Expr<$Output>;

            fn $method(self, rhs: Rhs) -> Self::Output {
                let $lhs = self;
                let $rhs = rhs.to_expr();
                $($body)*
            }
        }
    };
}

macro_rules! forward_value_to_expr_binary_op_trait {
    (
        #[generics($($generics:tt)*)]
        #[lhs_value($LhsValue:ty)]
        $Trait:ident::$method:ident
    ) => {
        impl<
                Rhs,
                $($generics)*
            > $Trait<Rhs> for $LhsValue
        where
            Expr<<$LhsValue as ToExpr>::Type>: $Trait<Rhs>,
        {
            type Output = <Expr<<$LhsValue as ToExpr>::Type> as $Trait<Rhs>>::Output;

            fn $method(self, rhs: Rhs) -> Self::Output {
                $Trait::$method(self.to_expr(), rhs)
            }
        }
    };
}

macro_rules! binary_op_bitwise {
    ($name:ident, $ty:ident, bool, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty,
                }
            }
        }

        /// intentionally only implemented for expressions, not rust's bool type,
        /// since that helps avoid using `==`/`!=` in hdl boolean expressions, which doesn't do
        /// what is usually wanted.
        impl $Trait for Expr<Bool> {
            type Output = Expr<Bool>;

            fn $method(self, rhs: Expr<Bool>) -> Expr<Bool> {
                $name { lhs: self, rhs }.to_expr()
            }
        }
    };
    ($name:ident, $ty:ident, $value:ident, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = UInt;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: UInt::new_dyn(Expr::ty(self.lhs).width().max(Expr::ty(self.lhs).width())),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn $Trait::$method(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> UInt {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            $Trait::$method
        }
    };
}

binary_op_bitwise!(BitAndU, UInt, UIntValue, BitAnd::bitand);
binary_op_bitwise!(BitAndS, SInt, SIntValue, BitAnd::bitand);
binary_op_bitwise!(BitAndB, Bool, bool, BitAnd::bitand);
binary_op_bitwise!(BitOrU, UInt, UIntValue, BitOr::bitor);
binary_op_bitwise!(BitOrS, SInt, SIntValue, BitOr::bitor);
binary_op_bitwise!(BitOrB, Bool, bool, BitOr::bitor);
binary_op_bitwise!(BitXorU, UInt, UIntValue, BitXor::bitxor);
binary_op_bitwise!(BitXorS, SInt, SIntValue, BitXor::bitxor);
binary_op_bitwise!(BitXorB, Bool, bool, BitXor::bitxor);

macro_rules! binary_op_add_sub {
    ($name:ident, $ty:ident, $value:ident, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        Expr::ty(self.lhs)
                            .width()
                            .max(Expr::ty(self.rhs).width())
                            .checked_add(1)
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn $Trait::$method(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            $Trait::$method
        }
    };
}

binary_op_add_sub!(AddU, UInt, UIntValue, Add::add);
binary_op_add_sub!(AddS, SInt, SIntValue, Add::add);
binary_op_add_sub!(SubU, UInt, UIntValue, Sub::sub);
binary_op_add_sub!(SubS, SInt, SIntValue, Sub::sub);

macro_rules! binary_op_mul {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        Expr::ty(self.lhs)
                            .width()
                            .checked_add(Expr::ty(self.rhs).width())
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn Mul::mul(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            Mul::mul
        }
    };
}

binary_op_mul!(MulU, UInt, UIntValue);
binary_op_mul!(MulS, SInt, SIntValue);

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DivU<LhsWidth: GenericConstUsize = UnknownSize> {
    pub lhs: Expr<UInt<LhsWidth>>,
    pub rhs: Expr<UInt>,
}

impl<LhsWidth: GenericConstUsize> ToExpr for DivU<LhsWidth> {
    type Type = UInt<LhsWidth>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DivU(DivU {
                lhs: Expr::as_dyn_int(self.lhs),
                rhs: self.rhs,
            })
            .intern(),
            __ty: Expr::ty(self.lhs),
        }
    }
}

impl_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
    fn Div::div(lhs: UInt<LhsWidth>, rhs: UInt<RhsWidth>) -> UInt<LhsWidth> {
        DivU {
            lhs,
            rhs: Expr::as_dyn_int(rhs),
        }
        .to_expr()
    }
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Div::div
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DivS {
    pub lhs: Expr<SInt>,
    pub rhs: Expr<SInt>,
}

impl ToExpr for DivS {
    type Type = SInt;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DivS(*self).intern(),
            __ty: SInt::new_dyn(
                Expr::ty(self.lhs)
                    .width()
                    .checked_add(1)
                    .expect("width too big"),
            ),
        }
    }
}

impl_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
    fn Div::div(lhs: SInt<LhsWidth>, rhs: SInt<RhsWidth>) -> SInt {
        DivS {
            lhs: Expr::as_dyn_int(lhs),
            rhs: Expr::as_dyn_int(rhs),
        }
        .to_expr()
    }
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Div::div
}

macro_rules! binary_op_rem {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(Expr::ty(self.lhs).width().min(Expr::ty(self.rhs).width())),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn Rem::rem(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            Rem::rem
        }
    };
}

binary_op_rem!(RemU, UInt, UIntValue);
binary_op_rem!(RemS, SInt, SIntValue);

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct BundleLiteral<T: BundleType = Bundle> {
    ty: T,
    field_values: Interned<[Expr<CanonicalType>]>,
}

impl<T: BundleType> BundleLiteral<T> {
    #[track_caller]
    pub fn new(ty: T, field_values: &[Expr<CanonicalType>]) -> Self {
        let fields = ty.fields();
        assert_eq!(fields.len(), field_values.len());
        for (&field, &field_value) in fields.iter().zip(field_values) {
            assert_eq!(
                field.ty,
                Expr::ty(field_value),
                "field's type doesn't match value's type: field name {:?}",
                field.name
            );
        }
        Self {
            ty,
            field_values: field_values.intern(),
        }
    }
    pub fn ty(self) -> T {
        self.ty
    }
    pub fn field_values(self) -> Interned<[Expr<CanonicalType>]> {
        self.field_values
    }
}

impl<T: BundleType> ToExpr for BundleLiteral<T> {
    type Type = T;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::BundleLiteral(BundleLiteral {
                ty: Bundle::new(&self.ty.fields()),
                field_values: self.field_values,
            })
            .intern(),
            __ty: self.ty,
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct EnumLiteral<T: EnumType = Enum> {
    ty: T,
    variant_index: usize,
    variant_value: Option<Expr<CanonicalType>>,
}

impl<T: EnumType> EnumLiteral<T> {
    #[track_caller]
    pub fn new(ty: T, variant_index: usize, variant_value: Option<Expr<CanonicalType>>) -> Self {
        let variants = ty.variants();
        assert!(variant_index < variants.len());
        assert_eq!(
            variants[variant_index].ty,
            variant_value.map(Expr::ty),
            "variant's type doesn't match value's type: variant name {:?}",
            variants[variant_index].name
        );
        Self {
            ty,
            variant_index,
            variant_value,
        }
    }
    pub fn ty(self) -> T {
        self.ty
    }
    pub fn variant_index(self) -> usize {
        self.variant_index
    }
    pub fn variant_value(self) -> Option<Expr<CanonicalType>> {
        self.variant_value
    }
}

impl<T: EnumType> ToExpr for EnumLiteral<T> {
    type Type = T;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::EnumLiteral(EnumLiteral {
                ty: Enum::new(&self.ty.variants()),
                variant_index: self.variant_index,
                variant_value: self.variant_value,
            })
            .intern(),
            __ty: self.ty,
        }
    }
}

macro_rules! impl_dyn_shl {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<UInt>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            #[track_caller]
            fn to_expr(&self) -> Expr<Self::Type> {
                let Some(pow2_rhs_width) = Expr::ty(self.rhs)
                    .width()
                    .try_into()
                    .ok()
                    .and_then(|v| 2usize.checked_pow(v))
                else {
                    panic!(
                        "dynamic left-shift amount's bit-width is too big, try casting the shift \
                        amount to a smaller bit-width before shifting"
                    );
                };
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        (pow2_rhs_width - 1)
                            .checked_add(Expr::ty(self.lhs).width())
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn Shl::shl(lhs: $ty<LhsWidth>, rhs: UInt<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }
    };
}

impl_dyn_shl!(DynShlU, UInt, UIntValue);
impl_dyn_shl!(DynShlS, SInt, SIntValue);

macro_rules! impl_dyn_shr {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name<LhsWidth: GenericConstUsize = UnknownSize> {
            pub lhs: Expr<$ty<LhsWidth>>,
            pub rhs: Expr<UInt>,
        }

        impl<LhsWidth: GenericConstUsize> ToExpr for $name<LhsWidth> {
            type Type = $ty<LhsWidth>;

            #[track_caller]
            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name($name {
                        lhs: Expr::as_dyn_int(self.lhs),
                        rhs: self.rhs,
                    })
                    .intern(),
                    __ty: Expr::ty(self.lhs),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn Shr::shr(lhs: $ty<LhsWidth>, rhs: UInt<RhsWidth>) -> $ty<LhsWidth> {
                $name {
                    lhs,
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }
    };
}

impl_dyn_shr!(DynShrU, UInt, UIntValue);
impl_dyn_shr!(DynShrS, SInt, SIntValue);

fn fixed_shr_width(lhs_width: usize, rhs: usize) -> Option<usize> {
    Some(lhs_width.saturating_sub(rhs).max(1))
}

macro_rules! binary_op_fixed_shift {
    ($name:ident, $ty:ident, $value:ident, $width_fn:path, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: usize,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        $width_fn(Expr::ty(self.lhs).width(), self.rhs).expect("width too big"),
                    ),
                }
            }
        }

        impl<LhsWidth: GenericConstUsize> $Trait<usize> for Expr<$ty<LhsWidth>> {
            type Output = Expr<$ty>;

            fn $method(self, rhs: usize) -> Self::Output {
                $name {
                    lhs: Expr::as_dyn_int(self),
                    rhs,
                }
                .to_expr()
            }
        }
    };
}

binary_op_fixed_shift!(FixedShlU, UInt, UIntValue, usize::checked_add, Shl::shl);
binary_op_fixed_shift!(FixedShlS, SInt, SIntValue, usize::checked_add, Shl::shl);
binary_op_fixed_shift!(FixedShrU, UInt, UIntValue, fixed_shr_width, Shr::shr);
binary_op_fixed_shift!(FixedShrS, SInt, SIntValue, fixed_shr_width, Shr::shr);

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Shl::shl
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Shl::shl
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Shr::shr
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Shr::shr
}

pub trait IntCmpExpr<Rhs: Type>: Type {
    fn cmp_eq(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_ne(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_lt(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_le(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_gt(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_ge(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
}

impl<Lhs: ToExpr, Rhs: ToExpr> IntCmp<Rhs> for Lhs
where
    Lhs::Type: IntCmpExpr<Rhs::Type>,
{
    fn cmp_eq(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_eq(self.to_expr(), rhs.to_expr())
    }
    fn cmp_ne(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_ne(self.to_expr(), rhs.to_expr())
    }
    fn cmp_lt(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_lt(self.to_expr(), rhs.to_expr())
    }
    fn cmp_le(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_le(self.to_expr(), rhs.to_expr())
    }
    fn cmp_gt(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_gt(self.to_expr(), rhs.to_expr())
    }
    fn cmp_ge(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_ge(self.to_expr(), rhs.to_expr())
    }
}

macro_rules! impl_compare_op {
    (
        $(#[width($LhsWidth:ident, $RhsWidth:ident)])?
        #[dyn_type($DynTy:ident)]
        #[to_dyn_type($lhs:ident => $dyn_lhs:expr, $rhs:ident => $dyn_rhs:expr)]
        #[type($Lhs:ty, $Rhs:ty)]
        $(
            struct $name:ident;
            fn $method:ident();
        )*
    ) => {
        $(#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$DynTy>,
            pub rhs: Expr<$DynTy>,
        }

        impl ToExpr for $name {
            type Type = Bool;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: Bool,
                }
            }
        })*

        impl$(<$LhsWidth: GenericConstUsize, $RhsWidth: GenericConstUsize>)? IntCmpExpr<$Rhs> for $Lhs {
            $(fn $method($lhs: Expr<Self>, $rhs: Expr<$Rhs>) -> Expr<Bool> {
                $name {
                    lhs: $dyn_lhs,
                    rhs: $dyn_rhs,
                }.to_expr()
            })*
        }
    };
}

impl_compare_op! {
    #[dyn_type(Bool)]
    #[to_dyn_type(lhs => lhs, rhs => rhs)]
    #[type(Bool, Bool)]
    struct CmpEqB; fn cmp_eq();
    struct CmpNeB; fn cmp_ne();
    struct CmpLtB; fn cmp_lt();
    struct CmpLeB; fn cmp_le();
    struct CmpGtB; fn cmp_gt();
    struct CmpGeB; fn cmp_ge();
}

impl_compare_op! {
    #[width(LhsWidth, RhsWidth)]
    #[dyn_type(UInt)]
    #[to_dyn_type(lhs => Expr::as_dyn_int(lhs), rhs => Expr::as_dyn_int(rhs))]
    #[type(UInt<LhsWidth>, UInt<RhsWidth>)]
    struct CmpEqU; fn cmp_eq();
    struct CmpNeU; fn cmp_ne();
    struct CmpLtU; fn cmp_lt();
    struct CmpLeU; fn cmp_le();
    struct CmpGtU; fn cmp_gt();
    struct CmpGeU; fn cmp_ge();
}

impl_compare_op! {
    #[width(LhsWidth, RhsWidth)]
    #[dyn_type(SInt)]
    #[to_dyn_type(lhs => Expr::as_dyn_int(lhs), rhs => Expr::as_dyn_int(rhs))]
    #[type(SInt<LhsWidth>, SInt<RhsWidth>)]
    struct CmpEqS; fn cmp_eq();
    struct CmpNeS; fn cmp_ne();
    struct CmpLtS; fn cmp_lt();
    struct CmpLeS; fn cmp_le();
    struct CmpGtS; fn cmp_gt();
    struct CmpGeS; fn cmp_ge();
}
